use crate::consts::*;
use crate::data::{BotType, PlayConfiguration};
use serde_with::{serde_as, DisplayFromStr, NoneAsEmptyString};

#[serde_as]
#[derive(serde::Serialize, serde::Deserialize, Debug, Clone, Eq, PartialEq)]
pub struct GameRules {
    #[serde_as(as = "DisplayFromStr")]
    pub map_width: usize,
    #[serde_as(as = "DisplayFromStr")]
    pub map_height: usize,
    #[serde_as(as = "DisplayFromStr")]
    pub boats_str: String,
    #[serde_as(as = "DisplayFromStr")]
    pub boats_can_touch: bool,
    #[serde_as(as = "DisplayFromStr")]
    pub player_continue_on_hit: bool,
    #[serde_as(as = "NoneAsEmptyString")]
    pub strike_timeout: Option<u64>,
    pub bot_type: BotType,
}

impl Default for GameRules {
    fn default() -> Self {
        Self::random_players_rules()
    }
}

impl GameRules {
    pub fn random_players_rules() -> Self {
        Self {
            map_width: MULTI_PLAYER_MAP_WIDTH,
            map_height: MULTI_PLAYER_MAP_HEIGHT,
            boats_str: MULTI_PLAYER_PLAYER_BOATS
                .iter()
                .map(usize::to_string)
                .collect::<Vec<_>>()
                .join(","),
            boats_can_touch: MULTI_PLAYER_BOATS_CAN_TOUCH,
            player_continue_on_hit: MULTI_PLAYER_PLAYER_CAN_CONTINUE_AFTER_HIT,
            strike_timeout: Some(30),
            bot_type: BotType::Smart,
        }
    }

    pub fn with_bot_type(mut self, t: BotType) -> Self {
        self.bot_type = t;
        self
    }

    pub fn with_player_continue_on_hit(mut self, c: bool) -> Self {
        self.player_continue_on_hit = c;
        self
    }

    pub fn with_strike_timeout(mut self, timeout: u64) -> Self {
        self.strike_timeout = Some(timeout);
        self
    }

    /// Set the list of boats for this configuration
    pub fn set_boats_list(&mut self, boats: &[usize]) {
        self.boats_str = boats
            .iter()
            .map(usize::to_string)
            .collect::<Vec<_>>()
            .join(",");
    }

    /// Get the list of boats for this configuration
    pub fn boats_list(&self) -> Vec<usize> {
        if self.boats_str.is_empty() {
            return vec![];
        }

        self.boats_str
            .split(',')
            .map(|s| s.parse::<usize>().unwrap_or_default())
            .collect()
    }

    /// Remove last boat
    pub fn remove_last_boat(&mut self) -> usize {
        let list = self.boats_list();
        self.set_boats_list(&list[0..list.len() - 1]);
        *list.last().unwrap()
    }

    /// Add a boat to the list of boats
    pub fn add_boat(&mut self, len: usize) {
        let mut list = self.boats_list();
        list.push(len);
        self.set_boats_list(&list);
    }

    /// Check game rules errors
    pub fn get_errors(&self) -> Vec<&str> {
        let config = PlayConfiguration::default();

        let mut errors = vec![];

        if self.map_width < config.min_map_width {
            errors.push("Map width is too small!");
        }

        if self.map_width > config.max_map_width {
            errors.push("Map width is too big!");
        }

        if self.map_height < config.min_map_height {
            errors.push("Map height is too small!");
        }

        if self.map_height > config.max_map_height {
            errors.push("Map height is too big!");
        }

        if self.boats_list().len() < config.min_boats_number {
            errors.push("There is not enough boats!");
        }

        if self.boats_list().len() > config.max_boats_number {
            errors.push("There are too many boats!");
        }

        for boat in self.boats_list() {
            if boat < config.min_boat_len || boat > config.max_boat_len {
                errors.push("A boat has an invalid length");
            }
        }

        if let Some(timeout) = self.strike_timeout {
            if timeout < config.min_strike_timeout {
                errors.push("Strike timeout is too short!");
            }

            if timeout > config.max_strike_timeout {
                errors.push("Strike timeout is too long!");
            }
        }

        errors
    }

    /// Check out whether these game rules are valid or not
    pub fn is_valid(&self) -> bool {
        self.get_errors().is_empty()
    }
}

#[cfg(test)]
mod test {
    use crate::data::GameRules;

    #[test]
    fn multi_players_config() {
        assert!(GameRules::random_players_rules().get_errors().is_empty());
    }
}
